/*
 *  (C) Copyright 2003 Pawel Parys <pp209216@zodiac.mimuw.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <vector>
#include <string>
#include <algorithm>
#include <dirent.h>
#include <stdio.h>
#include <ncurses.h>
#include <unistd.h>

using namespace std;

string nnrdata = "nnrdata/";

// ---------------- tools -----------------

template<class T> inline void insertV(std::vector<T> &v, const T &el)
{
	v.resize(v.size()+1, el);
}

template<class T> inline T &insertV(std::vector<T> &v)
{
	v.resize(v.size()+1);
	return v[v.size()-1];
}

template<class T> inline void removeV(vector<T> &v, int pos)
{
	v[pos]=v[v.size()-1];
	v.resize(v.size()-1);
}

template<class T> inline void removeVLast(vector<T> &v)
{
	v.resize(v.size()-1);
}

void fatalError(const char *str)
{
	printf("-------------------------- Fatal error --------------------------------\n%s\n", str);
	exit(1);
}

char tmpStr[16*1024*1024];

// ---------------- strtools --------------

inline int hasSufix(const char *str, const char *suffix)
{
	return strlen(str)>=strlen(suffix) && strcmp(str+strlen(str)-strlen(suffix), suffix)==0;
}

inline void removeNL(char *str)
{
	int l = strlen(str);
	if (l && str[l-1]=='\n') --l;
	if (l && str[l-1]=='\r') --l;	
	str[l] = 0;
}

// --------------- folderslist -----------------

vector<string> folders;
int curFolder;

void loadFoldersList()
{
	DIR *dir = opendir(nnrdata.c_str());
	folders.resize(0);
	for (;;)
	{
		dirent *res = readdir(dir);
		if (res == NULL)
			break;
		if (!hasSufix(res->d_name, ".dat"))
			continue;
		insertV(folders, string(res->d_name).substr(0, strlen(res->d_name)-4));
	}
	closedir(dir);
	FILE *f = fopen((nnrdata+"dir").c_str(), "rt");
	if (f==NULL) 
		return;
	char buf[1024];
	int a = 0;
	while (!feof(f))
	{
		char *x = fgets(buf,1024,f);
		if (x==NULL || a>=folders.size()) break;
		removeNL(buf);
		vector<string>::iterator it = find(folders.begin(), folders.end(), buf);
		if (it == folders.end())
			continue;
		string s = *it;
		*it = folders[a];
		folders[a] = s;
		++a;
	}
	fclose(f);
}

//---------------- message ----------------

struct Message
{
	string head, body, id, parent, title, author, dateStr;
	int statusByte;
};

#define F_UNREAD 1
#define F_UNAVAIL 2
#define F_FLAG0 4
#define F_DELETED 8
#define F_FLAG1 0x10
#define F_CLOSED 0x20

//---------------- dbase ------------------

void readS(FILE *f, string &str)
{
        if (feof(f)) { str.resize(0); return; }

	int len;
        fread(&len, sizeof(int), 1, f);
        
        len = fread(&tmpStr[0], 1, len+1, f);
	tmpStr[len] = 0;
	str = tmpStr;
}

char encodeStatusByte(char ch)
{
	switch (ch){
		case F_UNREAD: return 'n';
		case 0: return ' ';
		case F_UNAVAIL|F_UNREAD: return '-';
		case F_DELETED: return 'o';
		case 'n': return F_UNREAD;
		case ' ': return 0;
		case '-': return F_UNAVAIL|F_UNREAD;
		case 'o': return F_DELETED;
	};
	return ch;
}

void readStatusByte(FILE *f, Message &msg)
{
	string str;
	readS(f, str);
	msg.statusByte = str.empty() ? 0 : encodeStatusByte(str[0]);
}

void fastReadMessage(FILE *f, Message &msg)
{
	readS(f, msg.head);
	readS(f, msg.body);
        readS(f, msg.id);
        readS(f, msg.parent);
        readS(f, msg.title);
        readS(f, msg.author);
	readStatusByte(f, msg);
	readS(f, msg.dateStr);
}

//---------------- msglist ----------------

vector<Message> messages;
int msgNo;
Message *msg;

void loadMessagesList()
{
	string msgfile = nnrdata+folders[curFolder];
	FILE *f = fopen(msgfile.c_str(), "rb");
	messages.resize(0);
	if (f)
	{
		for(;;)
		{
			Message &msg = insertV(messages);
			fastReadMessage(f, msg);

			if (msg.id.empty()) break;
			if (msg.statusByte&F_DELETED)
				{ removeVLast(messages); continue; }
		}
		fclose(f);
	}
}

FILE *fout;

int pozy(const string &a, char ch)
{
        int i,l=a.length(),c=0;
        for(i=0;i<l;i++)
        {
                if(a[i]=='\"') c++;
                if((c%2)==0 && a[i]==ch) return i;
        }
        return 0;
}

string wytmail2(const string &a)
{
	int i;
	int j;
	int rl,rr;

	i=pozy(a,'@');
	if(!i) return "";
	for(j=i-1;j>=0;j--) if(a[j]==' ' || a[j]=='<') break;
	rl=j+1;
	for(j=i+1;j<a.size();j++) if(a[j]==' ' || a[j]=='>') break;
	rr=j-1;
	return a.substr(rl,rr-rl+1);
}

void correctAuthor()
{
//	printf("%s -> ", msg->author.c_str());
	msg->author = wytmail2(msg->author);
//	printf("%s\n", msg->author.c_str());
}

void correctDate()
{
//	printf("%s -> ", msg->dateStr.c_str());
	char dw[100]="   ", m[100], h[100], res[200];
	int d, y;
	sscanf(msg->dateStr.c_str(), " %c%c%c, %d %s %d %s", dw, dw+1, dw+2, &d, m, &y, h);
	sprintf(res, "%s %s %.2d %s %d", dw, m, d, h, y);
	msg->dateStr = res;
//	printf("%s\n", msg->dateStr.c_str());
}

void skipNL(string &str)
{
	int a=0, b=0; 
	while (a<str.length())
		if (str[a]=='\r') a++;
		else str[b++]=str[a++];
	str.resize(b);
}

int ilezal, koduj[50], mamzal, sent;
string zalacz[50], zalnazwa[50];
string bound;

int strpos(const string &a, const string &b)
{
	char *x = strstr(a.c_str(), b.c_str());
	return x==NULL?-1:x-a.c_str();
}

string wycpole(const string &or, const string &a)
{
	string res;
	int l = strpos(or, a);
	if (l>=0)
	{
		res = or.substr(l+a.length());
		res = res.substr(0, strpos(res, "\n"));
	}
	else res="";
	return res;
}

void wyslij()
{
	string head;
	head = head + "From: " + msg->author + "\n";
	head = head + "To: " + wycpole(msg->head, "Newsgroups: ") + "\n";
	head = head + "Subject: " + msg->title + "\n";
	head = head + "Date: " + msg->dateStr + "\n";
	head = head + "References: " + wycpole(msg->head, "References: ") + "\n";
	msg->head = head;
}

void robzalacznik()
{
	string w = "This message is in the MIME format and contains attachments.\n";
	w = w+"--"+bound+"\n";
	w = w+"Content-Type: text/plain; charset=\"iso-8859-2\"\n";
	w = w+"Content-Transfer-Encoding: 8bit\n\n";
	w = w+msg->body+"\n";
	for (int i = 0; i<ilezal; i++)
	{
		w = w+"--"+bound+"\n";
		if (koduj[i])
		{
			w = w+"Content-Type: application/octet-stream; name=\""+zalnazwa[i]+"\"\n";
			w = w+"Content-Transfer-Encoding: base64\n\n";
		}
		else
		{
			string rodzaj = (hasSufix(zalnazwa[i].c_str(), ".html") || hasSufix(zalnazwa[i].c_str(), ".htm"))
				? "html" : "text";
			w = w+"Content-Type: text/"+rodzaj+"; charset=\"iso-8859-2\"; name=\""+zalnazwa[i]+"\"\n";
			w = w+"Content-Transfer-Encoding: 8bit\n\n";
		}
		w = w+zalacz[i];
		w = w+"\n";
	}
	w = w+"\n--"+bound+"--\n";
	msg->body = w;
}

void naiso(string &a)
{
	char k2[101] = "󜟿ʣӌ";
	char k1[101] = "󶼿ʣӦ";
	for (int l = 0; l<a.length(); l++)
		for (int i = 0; i<16; i++)
			if (a[l]==k2[i])
			{
				a[l] = k1[i];
				break;
			}
}

void lowercase(string &str)
{
	for (int a = 0; a<str.length(); a++)
		if (str[a]>='A' && str[a]<='Z')
			str[a]+='a'-'A';
}

//do konca linii lub srednika, zamienia litery na male
//boundary=
string wycpoleb(string or, const string &a)
{
	string res = or;
	lowercase(or);
	int l = strpos(or, a);
	if (l>=0)
	{
		res = res.substr(l+a.length());
		l = strpos(res, "\n");
		int l2 = strpos(res, ";"); 
		if (l2>=0 && l2<l) l = l2;
		res = res.substr(0, l);
	}
	else
		res = "";
	return res;
}

void znajdz_zal()
{
	bound = wycpoleb(msg->head, "boundary=");
	if (!bound.empty() && bound[0]=='\"') bound = bound.substr(1, bound.length()-2);
	string c = msg->body;
	ilezal = mamzal = 0;
	if (!bound.empty() && strpos(c, bound)>=0)
	{
//		printf("Bound: %s\n", bound.c_str());
		int j = strpos(c, bound);
		msg->body = msg->body.substr(0, j);
		int dlb = bound.length();
		int n = -1;
		for(;;)
		{
			int k = strpos(c, bound);
			if (k<0) break;
			string d = c.substr(0, k);
			if (n>=0)
			{
				zalacz[n] = d.substr(strpos(d, "\n")+1);
				zalnazwa[n] = d.substr(0, strpos(d, "\n"));
				string xx = zalnazwa[n];
				int ll = strpos(xx, "<|\\>");
				if (ll>=0) { xx = xx.substr(0, ll); koduj[n]=1; } else koduj[n]=0;
				zalnazwa[n] = xx;
//				printf("zal: %s\nnazwa: %s\n\n", zalacz[n].c_str(), zalnazwa[n].c_str());
			}
			n++;
			c = c.substr(k+dlb+1);
		}
//		if (n>1 && (zalnazw[1].Pos(".html") || zalnazw[1].Pos(".htm")))
//			html=1;
		ilezal = n;
//		printf("Zalacznikow: %d\n", ilezal);
		mamzal = 1;
	}
}

void exportFolder()
{
//	if (curFolder!=1) return;
	if (folders[curFolder][0]!='[') return;
	sent = (strcmp(folders[curFolder].c_str(), "[Sent]")==0);
	printf("Exporting folder %s\n", folders[curFolder].c_str());
	loadMessagesList();
	fout = fopen(folders[curFolder].c_str(), "w");
	for (msgNo = 0; msgNo<messages.size(); msgNo++)
	{
		msg = &messages[msgNo];
		correctAuthor();
		correctDate();
		if (msg->head.length()<10) continue;
		fprintf(fout, "From %s %s\n", msg->author.c_str(), msg->dateStr.c_str());
		skipNL(msg->head);
		skipNL(msg->body);
		int x;
		if (sent)
			wyslij();
		else
		{
			for (x = 0; msg->head[x]!='\n'; x++); x++;
			msg->head = msg->head.substr(x);
		}
		znajdz_zal();
		naiso(msg->body);
		if (mamzal) robzalacznik();
		fprintf(fout, "%s\n", msg->head.c_str());
		fprintf(fout, "%s\n\n", msg->body.c_str());
	}
	fclose(fout);
}

// ------------- folderslist ------------

void exportDirectory()
{
	printf("Exporting directory %s\n", nnrdata.c_str());
	loadFoldersList();
	if (find(folders.begin(), folders.end(), "[Sent]")==folders.end())
	{
		loadFoldersList();
		if (find(folders.begin(), folders.end(), "[Sent]")==folders.end())
			fatalError("Can not create [Sent] folder");
	}
	for (curFolder = 0; curFolder<folders.size(); curFolder++)
		exportFolder();
}
// ----------------- nnrt --------------

int main()
{
	exportDirectory();
//	for (int a = 0; a<messages.size(); a++)
//		printf("%s %s\n", messages[a].author.c_str(), messages[a].title.c_str());
//	sleep(2);
	return 0;
}
